/* * *
 * sniffoid - Ukazka prace s libpcap. Zachytavani a parzovani paketu
 *            v urcenych sitich (= mereni prutoku dat).
 *
 * Verze: 0.1 alfa
 * Datum: 19.05. 2008
 *
 * Public domain
 *
 *
 *     Program snifuje z urceneho zarizeni (musi to byt ramce typu ethernet).
 * Jeho ucel je zobrazovat prutoky dat v jednotlivych sitich. Rozsah siti je
 * uveden bud v souborech, nebo si to prorgram parzuje ze standardniho vstupu.
 * 
 *     Jeho vystupem je adresa site a maska zony a pod nimi je uvedeno:
 *
 *     1. Kolikaty paket prave dorazil.
 *     2. Delka paketu + celkovy prenos.
 *     3. IP adresy ODKUD --> KAM
 *     4. Protokol, ktery je do IP zapouzdren.
 *        - zde rozeznavam 4 zakladni: TCP, UDP, ICMP, IP
 * 
 *     V souboru jsou zony uvedeny na samostatnych radcich (co novy radek, to
 * zona). Adresy se mohou zapisovat bud v dlouhe podobe (ip a maska jsou 
 * oddeleny mezerou), nebo v kratke (maska je uvedena za lomitkem):
 * 
 *     192.168.1.0 255.255.255.254
 *     1.2.3.4/5
 * 
 *     Vse ostatni je ignorovano; tedy az na specialni vyjimku: pokud misto
 * *cele* masky uvedeme cokoliv nesmyslneho (treba i abecedni znaky), je to
 * brano jako nulova maska! Jelikoz je to specialni pripad a nevedel jsem jak
 * se k tomu zachovat, nechal jsem to byt.
 *
 *     Program lze jednoduse rozsirit, ci predelat.
 *
 * Zdroje:
 *      [1] Tutorialy na http://www.tcpdump.org/
 *      [2] man pcap
 *      [3] http://www.google.com/
 * 
 * * */

#define _GNU_SOURCE

#include <netinet/in.h>
#include <arpa/inet.h>
#include <pcap.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "sniffoid.h"


static struct zony *prvni_zona, *zona;      /* Fronta zon */
static const char *prog_name;               /* Jmeno programu */


void help()
{
    fprintf(stderr, "Pouziti: %s [parametry] [soubor s pravidly ...].\n\n", prog_name);
    fprintf(stderr, "    Program zaznamenava prutok dat na urcenem zarizeni. "\
                    "(ethernet) Rozsah siti,\npro ktere bude merit, je zadan "\
                    "v souborech, nebo si je program nacita z stdin.\n\n");
    fprintf(stderr, "  -d dev    Zarizeni na kterem se bude sniffovat.\n");
    fprintf(stderr, "  -P n      Maximalini pocet prenesenych paketu.\n");
    fprintf(stderr, "  -p [01]   Vypnuti/zapnuti promiskuidniho modu.\n");
    fprintf(stderr, "  -h        Vypise tuto napovedu.\n\n");
}

void got_packet(u_char *args, const struct pcap_pkthdr *hdr, const u_char *pkt)
{
/*    struct ether_arp *ethernet = (struct ether_arp *) pkt; */
    struct ip_hdr *ip_cap;

    ip_cap = (struct ip_hdr *) (pkt + ETH_SIZE);
    zona = prvni_zona;

    while(zona->next)
    {
        if((zona->net.s_addr == GETNET(ip_cap->ip_src.s_addr, zona->mask.s_addr)) ||
           (zona->net.s_addr == GETNET(ip_cap->ip_dst.s_addr, zona->mask.s_addr)))
        {
            printf("  ------------------------------ \n");
            printf("ZONA:  << %s / ", inet_ntoa(zona->net));
            printf("%s >>\n", inet_ntoa(zona->mask));
            printf("     Paket: %lld\n", ++zona->pkt);
            printf("     Delka: %d B  (preneseno celkove: %lld B)\n", hdr->len, 
                   (zona->preneseno += hdr->len));
            printf("        Od: %s\n", inet_ntoa(ip_cap->ip_src));
            printf("       Kam: %s\n", inet_ntoa(ip_cap->ip_dst));
            printf("     Proto: ");

            switch(ip_cap->protocol)
            {
                case IPPROTO_TCP:
                    printf("TCP\n");
                    break;
                case IPPROTO_UDP:
                    printf("UDP\n");
                    break;
                case IPPROTO_ICMP:
                    printf("ICMP\n");
                    break;
                case IPPROTO_IP:
                    printf("IP\n");
                    break;
                default:
                    printf("neznamy\n");
                    break;
            }
        }

        zona = zona->next;
    }

}

struct zony *zparzuj_cfg(FILE *F_in, struct zony *zona, char *f_name)
{
    struct in_addr net_tmp;
    size_t line_len = 0;
    char *line = NULL;
    char *septemp;
    int n_tmp;

    while((getline(&line, &line_len, F_in)) > 0)
    {
        if(line[0] == '\n')
            continue;

        septemp = line;

        if((inet_aton(strsep(&septemp, " "), &net_tmp) != 0)) 
        {
            if(inet_aton(septemp, &(zona->mask)) != 0)
                zona->net.s_addr = GETNET(net_tmp.s_addr, zona->mask.s_addr);
        }
        else
        {
            septemp = line;

            if((inet_aton(strsep(&septemp, "/"), &net_tmp) != 0))
            {
                if(((n_tmp = strtol(septemp, NULL, 10)) < 0) || (n_tmp > 32))
                {
                    if(f_name)
                        fprintf(stderr, "%s: ", f_name);
                    fprintf(stderr, "Chybne zadana maska '%d'!\n", n_tmp);
                    continue;
                }
                zona->mask.s_addr = ntohl(PRFX2INT(n_tmp));
                zona->net.s_addr = GETNET(net_tmp.s_addr, zona->mask.s_addr);
            }
            else
            {
                if(f_name)
                    fprintf(stderr, "%s: ", f_name);
                fprintf(stderr, "Vyraz '%s' se nepodarilo zparzovat!\n", line);
                continue;
            }
        }
        /*            printf("net:  %s\n", inet_ntoa(zona->net));
                      printf("mask: %s\n\n", inet_ntoa(zona->mask));
                      */
        if(!(zona->next = (struct zony *) malloc(sizeof(struct zony))))
        {
            perror("malloc()");
            exit(EXIT_FAILURE);
        }

        zona = zona->next;
        zona->pkt = zona->preneseno = 0;

    }

    return zona;
}

int main(int argc, char **argv)
{
    char errbuf[PCAP_ERRBUF_SIZE];     /* Kam bude pcap zapisovat chyby */
    char filtr[] = "ip";
    char *dev = NULL;                  /* Jmeno device */
    uint32_t n_pkts = 0;               /* Maximalni pocet prenesenych paketu */
    uint16_t promisc = 1;              /* Ne/pomiskuidni mod */
    int opt;
    bpf_u_int32 my_ip, my_mask;
    pcap_t *handle;                    /* Jiz otevrene zarizeni */
    FILE *F_in;                        /* Vstupni datovy proud */
    struct bpf_program fp;             /* "Zkompilovany" filtrovaci vyraz */


    prog_name = basename(argv[0]);

    while((opt = getopt(argc, argv, "d:p:P:r:h")) != -1)
    {
        switch(opt)
        {
            case 'P': /* Max pocet prenesenych paketu */
                n_pkts = strtol(optarg, NULL, 10);
                break;
            case 'd': /* Device */
                dev = strdup(optarg);
                break;
            case 'p': /* Zapnuti/vypnuti promiskuidniho modu */
                promisc = strtol(optarg, NULL, 10);
                break;
            case 'h':   /* Help */
                help();
                exit(EXIT_SUCCESS);
            case ':':   /* Chybi hodnota parametru */
            case '?':   /* Neznama opsna */
            default:
                fprintf(stderr, "Chybne zadany parametr!\n");
                exit(EXIT_FAILURE);
        }
    }
#define ZBL 1
#if ZBL
    /* 
     * Zkusi device najit
     */
    if((dev == NULL) && ((dev = pcap_lookupdev(errbuf)) == NULL)) {
        fprintf(stderr, "Vychozi zarizeni: %s\n", errbuf);
        return 1;
    }

    /* Zkus ziskat IP a MASKU zarizeni
     */
    if (pcap_lookupnet(dev, &my_ip, &my_mask, errbuf) == -1) {
        fprintf(stderr, "Nelze ziskat ip/masku ze zarizeni '%s' (%s)\n", dev, errbuf);
        my_ip = my_mask = 0;
    }

    /* Nyni otevreme zarizeni pro sniffing
     */
    if((handle = pcap_open_live(dev, BUFSIZ, promisc, 1000, errbuf)) == NULL) {
        fprintf(stderr, "Zarizeni '%s' nelze otevrit: %s\n", dev, errbuf);
        return 1;
    }

    /* Sniffuji pouze na ethernetu! (nic jineho nemam k dispozici)
     */
    if (pcap_datalink(handle) != DLT_EN10MB) {
        fprintf(stderr, "%s neni Ethernet!\n", dev);
        return 1;
    }
#endif

    if(!(zona = (struct zony *) malloc(sizeof(*zona)))) {
        perror("malloc()");
        return -1;
    }

    zona->pkt = zona->preneseno = 0;
    prvni_zona = zona;

    if(optind == argc)
    {
        fprintf(stderr, "Nacitam z `stdin' a zapisuju na `stdout':\n");
        zona = zparzuj_cfg(stdin, zona, argv[optind]);
    }
    else
    {
        for(; optind < argc; optind++)
        {

            if(!(F_in = fopen(argv[optind], "rt")))
            {
                fprintf(stderr, "fopen: ");
                perror(argv[optind]);
                continue;
            }

            zona = zparzuj_cfg(F_in, zona, argv[optind]);

            fclose(F_in);
        }
    }

    zona->net.s_addr = -1;
    zona->mask.s_addr = 0;
    zona->next = NULL;
    zona = prvni_zona;

    printf("ZONY (init):\n");

    while(zona->next)
    {
        
        printf("  net:  %s\n", inet_ntoa(zona->net));
        printf("  mask: %s\n", inet_ntoa(zona->mask));
        printf("  pkt:  %lld\n", zona->pkt);
        printf("  len:  %lld\n\n", zona->preneseno);
        zona = zona->next;
    }

#if ZBL
    if(pcap_compile(handle, &fp, filtr, 0, my_ip) == -1) {
        fprintf(stderr, "Chyba pri parzovani filtru (%s): %s\n", filtr, pcap_geterr(handle));
        return 1;
    }

    if(pcap_setfilter(handle, &fp) == -1) {
        fprintf(stderr, "Nemohu nastavit filtr (%s): %s\n", filtr, pcap_geterr(handle));
        return 1;
    }

    pcap_loop(handle, n_pkts, got_packet, NULL);

    pcap_freecode(&fp);
    pcap_close(handle);
#endif

    return 0;
}

